#ifndef DICTIONARY_GUARD
#define DICTIONARY_GUARD

#include <assert.h>
#include <limits.h>
#include <iostream>
using namespace std;


typedef unsigned short height_t;
typedef unsigned short refcount_t;

enum STATUS { NOT_FOUND, FOUND };

static const refcount_t maxRefCount = USHRT_MAX;

template <class Key_t, class Datum_t>
class DictNode_T;

template <class Key_t, class Datum_t>
class Dictionary_T {
	friend int keycompare(Key_t k1, Key_t k2);
	friend ostream& operator<< (ostream & out, const Dictionary_T<Key_t, Datum_t> &d);
public:
	Dictionary_T();                                  // Default constructor
	Dictionary_T(DictNode_T<Key_t, Datum_t> *n);                       // Constructor
	Dictionary_T(Key_t k, Datum_t v, Dictionary_T<Key_t, Datum_t> l, Dictionary_T<Key_t, Datum_t> r); // Constructor
	Dictionary_T(Key_t k, Datum_t v);                   // Constructor
	Dictionary_T(const Dictionary_T<Key_t, Datum_t> &d);               // Copy constructor
	~Dictionary_T();                                 // Destructor
	Dictionary_T<Key_t, Datum_t> & operator= (const Dictionary_T<Key_t, Datum_t> &d);  // Assignment operator
	bool NoEntries();
	Key_t Key();
	Datum_t Datum();
	Dictionary_T<Key_t, Datum_t> Left();
	Dictionary_T<Key_t, Datum_t> Right();
	Dictionary_T<Key_t, Datum_t> RightMost();
	height_t Height();
	Dictionary_T<Key_t, Datum_t> Insert(Key_t k, Datum_t v);
	Dictionary_T<Key_t, Datum_t> Delete(Key_t k);
	STATUS Lookup(Key_t k, Datum_t &v);

	DictNode_T<Key_t, Datum_t> *myEntries;

	class iterator{
		public:
			iterator *Next() {
                iterator *prevEntry;
                iterator *iter=NULL;

                this->rightDone=true;
                prevEntry=this;
                DictNode_T<Key_t, Datum_t> *curNode=myEntry->right.myEntries;
                while(curNode) {
                    iter= new iterator();
                    iter->parent=prevEntry;
                    iter->myEntry=curNode;
                    iter->rightDone=false;
                    prevEntry=iter;
                    curNode=curNode->left.myEntries;
                }
                if(iter) {
                    return iter;
                }
                else {
                    // No right child
					iterator *temp;
                    iter=this->parent;
					while(iter && iter->rightDone) {
                        temp=iter;
                        iter=iter->parent;
                        delete temp;
                    }
                    return iter;
                }
            }
			const Key_t &Key() {
				return myEntry->key;
			}
			const Datum_t &Datum() {
				return myEntry->datum;
			}
		private:
			bool rightDone;		//Have gone down the right tree?
			iterator *parent;
			DictNode_T<Key_t, Datum_t> *myEntry;
			friend class Dictionary_T<Key_t, Datum_t>;
	};

	// Iterators
    typename Dictionary_T<Key_t, Datum_t>::iterator *First();
 private:

	Dictionary_T<Key_t, Datum_t> *iterParent;

	void Print(ostream & out, unsigned int depth) const;
	Dictionary_T<Key_t, Datum_t> Bal(Key_t k, Datum_t d, Dictionary_T<Key_t, Datum_t> l, Dictionary_T<Key_t, Datum_t> r);
};

template <class Key_t, class Datum_t>
class DictNode_T {
 public:
	DictNode_T(Key_t k, Datum_t d, Dictionary_T<Key_t, Datum_t> l, Dictionary_T<Key_t, Datum_t> r, height_t h);
	DictNode_T(Key_t k, Datum_t d, Dictionary_T<Key_t, Datum_t> l, Dictionary_T<Key_t, Datum_t> r);
	DictNode_T(Key_t k, Datum_t d);
	~DictNode_T<Key_t, Datum_t>();
	void IncrRef();
	void DecrRef();

 private:
	Key_t key;
	Datum_t datum;
	Dictionary_T<Key_t, Datum_t> left, right;
	height_t height;
	refcount_t refCount;
	friend class Dictionary_T<Key_t, Datum_t>;
	friend class Dictionary_T<Key_t, Datum_t>::iterator;
};

#endif
